"""
Client example for text classification active learning tasks.
@author: huangyz0918 (huangyz0918@gmail.com)
@date: 22/07/2022
"""
import time
from alaas.client import Client

if __name__ == '__main__':
    client = Client('http://0.0.0.0:8081')
    text_list = [
        'The movie itself was to me a huge disappointment.',
        'My name is Wolfgang and I live in Berlin',
        'My name is Wolfgang and I live in Berlin',
        'My name is Wolfgang and I live in Berlin',
        'My name is Wolfgang and I live in Berlin'
    ]
    start_time = time.time()
    queries = client.query_by_text(text_list, budget=3)
    end_al_time = time.time()
    print(queries)
    print(f"Latency: {end_al_time - start_time}")



"""
The basic server of ALaaS.
@author huangyz0918 (huangyz0918@gmail.com)
@date 20/07/2022
"""
from jina import Flow

from alaas.server.util import ConfigManager
from alaas.server.executors import TorchALWorker


class Server:
    """
    Server: Server Class for Active Learning Services.
    """
    @staticmethod
    def start(proto='http',
                port=8081, 
                host='0.0.0.0', 
                job_name='default_app', 
                model_hub='pytorch/vision:v0.10.0', 
                model_name='resnet18', 
                device='cpu', 
                strategy='LeastConfidence', 
                batch_size=1, 
                replica=1, 
                tokenizer=None, 
                transformers_task=None
            ):
        # start with an image classification application (with PyTorch resnet18) as the example.
        Flow(protocol=proto, port=port, host=host) \
            .add(name=job_name,
                 uses=TorchALWorker,
                 uses_with={
                     'model_name': model_name,
                     'model_repo': model_hub,
                     'device': device,
                     'strategy': strategy,
                     'minibatch_size': batch_size,
                     'tokenizer_model': tokenizer,
                     'task': transformers_task
                 },
                 replicas=replica) \
            .start()


    @staticmethod
    def start_by_config(config_path):
        # TODO: multi-worker support
        cfg_manager = ConfigManager(config_path)
        _host = cfg_manager.al_worker.host
        _port = cfg_manager.al_worker.port
        _replica = cfg_manager.al_worker.replicas
        _name = cfg_manager.name
        _proto = cfg_manager.al_worker.protocol

        # Active learning executor parameters.
        _strategy = cfg_manager.strategy.type.value
        _model_hub = cfg_manager.strategy.model.hub
        _model_name = cfg_manager.strategy.model.name
        _device = cfg_manager.strategy.model.device
        _batch_size = cfg_manager.strategy.model.batch_size
        # only for text data/model.
        _tokenizer = cfg_manager.strategy.model.tokenizer
        _task = cfg_manager.strategy.model.task

        Flow(protocol=_proto, port=_port, host=_host) \
            .add(name=_name,
                 uses=TorchALWorker,
                 uses_with={
                     'model_name': _model_name,
                     'model_repo': _model_hub,
                     'device': _device,
                     'strategy': _strategy,
                     'minibatch_size': _batch_size,
                     'tokenizer_model': _tokenizer,
                     'task': _task
                 },
                 replicas=_replica) \
            .start()



"""
Toolkit for data operations.
@author huangyz0918 (huangyz0918@gmail.com)
@date 28/05/2022
"""

import uuid
import hashlib
import sqlite3
import boto3
import urllib.request
import numpy as np
from abc import ABC, abstractmethod


class Downloader(ABC):
    """
    Downloader: the base data downloader class.
    """

    @abstractmethod
    def download(self, filename, **kwargs):
        pass


class UrlDownloader(Downloader):
    """
    Url Downloader: the data downloader of given urls.
    """

    def download(self, filename, url=None):
        if url:
            urllib.request.urlretrieve(url, filename)
            return filename
        else:
            return None


class S3Downloader(Downloader):
    """
    S3 Downloader: the data downloader of AWS S3.
    """

    def __init__(self, aws_access_key_id, aws_secret_access_key, bucket_name):
        session = boto3.session.Session()
        self.bucket = session.resource('s3',
                                       aws_access_key_id=aws_access_key_id,
                                       aws_secret_access_key=aws_secret_access_key).Bucket(bucket_name)

    def download(self, filename, object_name=None):
        self.bucket.download_file(object_name, filename)
        return filename


class DBManager:
    """
    DBManager: A manager for the database (SQLite) managements.
    """

    def __init__(self, db_file):
        """
        Create a database connection to a SQLite database.
        :param db_file: the database file location.
        """
        try:
            sqlite3.register_adapter(np.ndarray, adapt_array)
            sqlite3.register_converter("ARRAY", convert_array)
            self.conn = sqlite3.connect(db_file, detect_types=sqlite3.PARSE_DECLTYPES, check_same_thread=False)
            self.cursor = self.conn.cursor()
            self.cursor.execute("CREATE TABLE IF NOT EXISTS AL_POOL (MD5 TEXT, UUID TEXT, DATA TEXT, INFER ARRAY);")
        except Exception as e:
            print(e)

    def close_connection(self):
        """
        Close the connection of SQLite.
        """
        self.conn.close()

    def connect(self, db_file):
        """
        Establish the connection of SQLite.
        :param db_file: the database file location.
        """
        self.conn = sqlite3.connect(db_file)

    def insert_record(self, data_pth, infer_result=None, skip=True):
        """
        Insert the record to SQLite.
        :param data_pth: the path of inference request file.
        :param skip: skip the file with the same MD5.
        :param infer_result: the inference result according to the input data.
        """
        random_uuid = str(uuid.uuid4())
        file_id = hashlib.md5(data_pth.encode('utf-8')).hexdigest()
        if skip:
            if not self.check_row(file_id):
                self.cursor.execute("INSERT INTO AL_POOL VALUES (?, ?, ?, ?);",
                                    (file_id, random_uuid, data_pth, infer_result))
        else:
            self.cursor.execute("INSERT INTO AL_POOL VALUES (?, ?, ?, ?);",
                                (file_id, random_uuid, data_pth, infer_result))

        self.conn.commit()

    def update_inference(self, data_uuid, infer_result):
        """
        Update the inference result according to the given UUID.
        """
        self.cursor.execute("UPDATE AL_POOL SET INFER = ? WHERE UUID = ?", (infer_result, data_uuid))
        self.conn.commit()

    def update_inference_md5(self, data_md5, infer_result):
        """
        Update the inference result according to the given MD5.
        """
        self.cursor.execute("UPDATE AL_POOL SET INFER = ? WHERE MD5 = ?", (infer_result, data_md5))
        self.conn.commit()

    def read_records(self, with_infer=False):
        """
        Read all records from the current database.
        """
        if with_infer:
            return self.cursor.execute("SELECT MD5, UUID, DATA, INFER FROM AL_POOL").fetchall()
        else:
            return self.cursor.execute("SELECT MD5, UUID, DATA FROM AL_POOL").fetchall()

    def get_rows(self, inferred=True):
        """
        Get all the rows with inference results.
        """
        if inferred:
            return self.cursor.execute("SELECT MD5, UUID, DATA, INFER FROM AL_POOL WHERE INFER IS NOT NULL").fetchall()
        else:
            return self.cursor.execute("SELECT MD5, UUID, DATA, INFER FROM AL_POOL WHERE INFER IS NULL").fetchall()

    def check_row(self, md5):
        """
        Check is a row exists in the dataset by MD5.
        """
        result = self.cursor.execute("SELECT EXISTS(SELECT 1 FROM AL_POOL WHERE MD5 = ?);", (md5,)).fetchall()
        return result[0][0] != 0


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def adapt_array(arr):
    """
    Adapt the numpy.array to SQLite.
    """
    return arr.tobytes()


def convert_array(text):
    """
    Convert the SQLite bytes to numpy.array.
    """
    return np.frombuffer(text)



"""
The active learning model updater of PyTorch models.
@author huangyz0918 (huangyz0918@gmail.com)
@date 10/08/2022
"""
import torch
import torch.jit

import os
import warnings
from pathlib import Path
from multiprocessing.pool import ThreadPool
from jina import Executor, DocumentArray, requests

from alaas.server.util import DBManager


class TorchCLWorker(Executor):
    """
    TorchCLWorker: the backend model updating worker class of PyTorch models.
    """

    def __init__(
            self,
            device: str = None,
            minibatch_size: int = 128,
            num_worker_preprocess: int = 4,
            data_home: str = None,
            transform=None,
            *args,
            **kwargs,
    ):
        """
        TODO: add river-torch (https://github.com/online-ml/river-torch) or implement an online learning module here.
        """
        super().__init__(*args, **kwargs)

        self._transform = transform
        self._minibatch_size = minibatch_size
        self._thread_pool = ThreadPool(processes=num_worker_preprocess)

        if data_home is None:
            self._data_home = str(Path.home()) + "/.alaas/"
            Path(self._data_home).mkdir(parents=True, exist_ok=True)

        self._db_manager = DBManager(self._data_home + 'cache.db')
        self._data_pool = self._db_manager.read_records()
        self._data_processed = self._db_manager.get_rows()
        self._data_not_processed = self._db_manager.get_rows(inferred=False)

        if not device:
            self._device = 'cuda' if torch.cuda.is_available() else 'cpu'
        else:
            self._device = device

        if not self._device.startswith('cuda') and (
                'OMP_NUM_THREADS' not in os.environ
                and hasattr(self.runtime_args, 'replicas')
        ):
            replicas = getattr(self.runtime_args, 'replicas', 1)
            num_threads = max(1, torch.get_num_threads() // replicas)
            if num_threads < 2:
                warnings.warn(
                    f'Too many replicas ({replicas}) vs too few threads {num_threads} may result in '
                    f'sub-optimal performance.'
                )

                torch.set_num_threads(max(num_threads, 1))
                torch.set_num_interop_threads(1)

    @requests(on='/update')
    def update(self, docs: DocumentArray, parameters, **kwargs):
        """
        Active learning push function. This function is designed for large-scale input data/uris,
        the data will be fetched first and store temporary in the centre AL server.
        @param docs: the uploaded/targeted data, can be blob/uri for CV tasks or text for NLP tasks.
        @param parameters: the parameters include the active learning budget, etc.
        @param kwargs: the kwargs for the backend server.
        @return: the queried data in uris/blobs/texts.
        """
        index_pths = []
        # TODO: update the active learner at scale.
        return DocumentArray(index_pths)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Author: Li Yuanming
Author: huangyz0918 (huangyz0918@gmail.com)
Email: yuanmingleee@gmail.com
Date: May 23, 2022
"""
from abc import ABC
from enum import Enum
from typing import Union

from alaas.types.models.al_model import ALModelBase
from alaas.types.models.utils import TypeCheckMixin


class ALStrategyType(Enum):
    """Enum of supported AL strategy"""
    RANDOM_SAMPLING = 'RandomSampling'
    LEAST_CONFIDENCE = 'LeastConfidence'
    MARGIN_CONFIDENCE = 'MarginConfidence'
    RATIO_CONFIDENCE = 'RatioConfidence'
    ENTROPY_SAMPLING = 'EntropySampling'
    KMEANS_SAMPLING = 'KMeansSampling'
    BADGE_SAMPLING = 'BadgeSampling'
    BAYESIAN_DISAGREEMENT = 'BayesianDisagreement'
    MEAN_STD = 'MeanSTDSampling'
    VAR_RATIO = 'VarRatioSampling'


class ALStrategyBase(TypeCheckMixin[ALStrategyType], ABC):
    pass


class RandomSamplingConfig(ALStrategyBase):
    model: ALModelBase

    __required_type__ = ALStrategyType.RANDOM_SAMPLING


class LeastConfidenceConfig(ALStrategyBase):
    model: ALModelBase

    __required_type__ = ALStrategyType.LEAST_CONFIDENCE


class MarginConfidenceConfig(ALStrategyBase):
    model: ALModelBase

    __required_type__ = ALStrategyType.MARGIN_CONFIDENCE


class RatioConfidenceConfig(ALStrategyBase):
    model: ALModelBase

    __required_type__ = ALStrategyType.RATIO_CONFIDENCE


class EntropySamplingConfig(ALStrategyBase):
    model: ALModelBase

    __required_type__ = ALStrategyType.ENTROPY_SAMPLING


class KMeansSamplingConfig(ALStrategyBase):
    model: ALModelBase

    __required_type__ = ALStrategyType.KMEANS_SAMPLING


class BayesianDisagreementConfig(ALStrategyBase):
    model: ALModelBase

    __required_type__ = ALStrategyType.BAYESIAN_DISAGREEMENT


class BadgeSamplingConfig(ALStrategyBase):
    model: ALModelBase

    __required_type__ = ALStrategyType.BADGE_SAMPLING


class MeanSTDSamplingConfig(ALStrategyBase):
    model: ALModelBase

    __required_type__ = ALStrategyType.MEAN_STD


class VarRatioSamplingConfig(ALStrategyBase):
    model: ALModelBase

    __required_type__ = ALStrategyType.VAR_RATIO


ALStrategyConfigUnion = Union[
    RandomSamplingConfig,
    LeastConfidenceConfig,
    MarginConfidenceConfig,
    RatioConfidenceConfig,
    EntropySamplingConfig,
    KMeansSamplingConfig,
    BayesianDisagreementConfig,
    BadgeSamplingConfig,
    MeanSTDSamplingConfig,
    VarRatioSamplingConfig
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Author: Li Yuanming
Email: yuanmingleee@gmail.com
Date: May 23, 2022
"""
from typing import Generic, TypeVar

from pydantic import validator
from pydantic.generics import GenericModel

RequiredType = TypeVar('RequiredType')


class TypeCheckMixin(GenericModel, Generic[RequiredType]):
    """For auto detecting configuration class by set value of :code:`type`.
    """
    type: RequiredType

    __required_type__: RequiredType

    @validator('type')
    def check_type(cls, required_type: RequiredType) -> RequiredType:
        """
        Checks type value provided is the same as the required value.
        This is to generate validator for check :code:`type` field of subclasses of Generic typpe :class:`RequiredType`.
        """
        if required_type != cls.__required_type__:
            raise ValueError(f'Expected {cls.__required_type__} but got {required_type}')
        return required_type



"""
Server example for image classification active learning tasks.
@author: huangyz0918 (huangyz0918@gmail.com)
@date: 01/05/2022
"""

from alaas.server import Server

SERVER_CONFIG = './resnet18.yml'

if __name__ == '__main__':
    # start the server by an input configuration file.
    # Server.start_by_config(SERVER_CONFIG)

    # start the server by parameters. 
    Server.start()



"""
Server example for text classification active learning tasks.
@author: huangyz0918 (huangyz0918@gmail.com)
@date: 23/07/2022
"""

from alaas.server import Server

SERVER_CONFIG = './distilbert_base.yml'

if __name__ == '__main__':
    # start the server by an input configuration file.
    # Server.start_by_config(SERVER_CONFIG)

    Server.start(model_hub="huggingface/pytorch-transformers", 
                    model_name="distilbert-base-uncased",
                    tokenizer="distilbert-base-uncased",
                    transformers_task="text-classification",
                    strategy="RandomSampling"
                )



"""
Random sampling methods.
@author huangyz0918 (huangyz0918@gmail.com)
@date 06/05/2022
"""

import numpy as np
from .base import Strategy


class RandomSampling(Strategy):
    """
    Randomly Selected the query samples.
    """

    def __init__(self, path_mapping, n_drop):
        super(RandomSampling, self).__init__(path_mapping, n_drop)

    def query(self, n, embedding=None):
        self.check_query_num(n)
        _path_list = np.array(self.path_mapping)
        return _path_list[np.random.choice(self.pool_size, int(n), replace=False)]



"""
Client example for image classification active learning tasks.
@author: huangyz0918 (huangyz0918@gmail.com)
@date: 01/05/2022
"""
import time
from alaas.client import Client


if __name__ == '__main__':
    client = Client('http://0.0.0.0:8081')
    remote_file_list = [
        'https://www.cs.toronto.edu/~kriz/cifar-10-sample/airplane1.png',
        'https://www.cs.toronto.edu/~kriz/cifar-10-sample/airplane2.png',
        'https://www.cs.toronto.edu/~kriz/cifar-10-sample/airplane3.png',
        'https://www.cs.toronto.edu/~kriz/cifar-10-sample/airplane4.png',
        'https://www.cs.toronto.edu/~kriz/cifar-10-sample/airplane5.png',
        'https://www.cs.toronto.edu/~kriz/cifar-10-sample/airplane6.png'
    ]
    start_time = time.time()
    queries = client.query_by_uri(remote_file_list, budget=3)
    end_al_time = time.time()
    print(queries)
    print(f"Latency: {end_al_time - start_time}")



"""
Data processing and augmentation toolkit
For image data.
@author huangyz0918 (huangyz0918@gmail.com)
@date 20/07/2022
"""

import cv2
import numpy as np
from torchvision import transforms

img_transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Resize(256),
    transforms.CenterCrop(224),
    transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
])


def load_image_data_as_np(image_path):
    return np.array([img_transform(cv2.imread(image_path)).numpy()], dtype=np.float32)


def load_images_data_as_np(data_pool):
    img_list = []
    img_md5_list = []
    img_uuid_list = []
    for data in data_pool:
        img_md5_list.append(data[0])
        img_uuid_list.append(data[1])
        img_list.append(img_transform(cv2.imread(data[2])).numpy())
    img_list = np.array(img_list, dtype=np.float32)
    return img_md5_list, img_uuid_list, img_list



"""
Uncertainty based active learning methods.
@author huangyz0918 (huangyz0918@gmail.com)
@date 28/05/2022
"""

import numpy as np
from .base import Strategy


class LeastConfidence(Strategy):
    """
    Least Confidence Sampling.
    Reference: A Sequential Algorithm for Training Text Classifiers (1994).
    (https://arxiv.org/pdf/cmp-lg/9407020)
    """

    def __init__(self, path_mapping, n_drop):
        super(LeastConfidence, self).__init__(path_mapping, n_drop)

    def query(self, n, embeddings=None):
        self.check_query_num(n)
        _path_list = np.array(self.path_mapping)
        if self.n_drop:
            embeddings = embeddings.mean(0)
        _uncertainties = np.amax(embeddings, axis=1)
        return _path_list[_uncertainties.argsort()[:int(n)]]


class MarginConfidence(Strategy):
    """
    Margin of Confidence Sampling.
    Reference: Active Hidden Markov Models for Information Extraction (2001).
    (https://link.springer.com/chapter/10.1007/3-540-44816-0_31)
    """

    def __init__(self, path_mapping, n_drop):
        super(MarginConfidence, self).__init__(path_mapping, n_drop)

    def query(self, n, embeddings=None):
        self.check_query_num(n)
        _path_list = np.array(self.path_mapping)
        if self.n_drop:
            embeddings = embeddings.mean(0)
        _probs_sorted = -np.sort(-np.array(embeddings))
        _difference = _probs_sorted[:, 0] - _probs_sorted[:, 1]
        return _path_list[_difference.argsort()[:int(n)]]


class RatioConfidence(Strategy):
    """
    Ratio of Confidence Sampling.
    Reference: Active learning literature survey. (https://minds.wisconsin.edu/handle/1793/60660)
    """

    def __init__(self, path_mapping, n_drop):
        super(RatioConfidence, self).__init__(path_mapping, n_drop)

    def query(self, n, embeddings=None):
        self.check_query_num(n)
        _path_list = np.array(self.path_mapping)
        if self.n_drop:
            embeddings = embeddings.mean(0)
        _probs_sorted = -np.sort(-np.array(embeddings))
        _difference = _probs_sorted[:, 0] / _probs_sorted[:, 1]
        return _path_list[_difference.argsort()[:int(n)]]


class EntropySampling(Strategy):
    """
    Entropy Sampling.
    Reference: Active learning literature survey. (https://minds.wisconsin.edu/handle/1793/60660)
    """

    def __init__(self, path_mapping, n_drop):
        super(EntropySampling, self).__init__(path_mapping, n_drop)

    def query(self, n, embeddings=None):
        self.check_query_num(n)
        _path_list = np.array(self.path_mapping)
        if self.n_drop:
            embeddings = embeddings.mean(0)
        _entropy = (embeddings * np.log(embeddings)).sum(1)
        return _path_list[_entropy.argsort()[:int(n)]]


class BayesianDisagreement(Strategy):
    """
    Bayesian Active Learning Disagreement.
    Reference: https://arxiv.org/abs/1703.02910
    """

    def __init__(self, path_mapping, n_drop):
        super(BayesianDisagreement, self).__init__(path_mapping, n_drop)

    def query(self, n, embeddings=None):
        self.check_query_num(n)
        _path_list = np.array(self.path_mapping)

        if self.n_drop:
            embeddings_mean = embeddings.mean(0)
            _entropy = (-embeddings_mean * np.log(embeddings_mean)).sum(1)
            _entropy_mean = (-embeddings * np.log(embeddings)).sum(2).mean(0)

            _uncertainties = _entropy_mean - _entropy
            return _path_list[_uncertainties.argsort()[:int(n)]]
        else:
            raise ValueError("strategy {BayesianDisagreement} requires to set the parameter {n_drop} > 1")


class MeanSTDSampling(Strategy):
    """
    Mean Standard Deviation Sampling.
    Reference: Semantic segmentation of small objects and modeling of uncertainty in urban remote
                sensing images using deep convolutional neural networks, CVPR, 2016
                (https://ieeexplore.ieee.org/document/7789580)
    """

    def __init__(self, path_mapping, n_drop):
        super(MeanSTDSampling, self).__init__(path_mapping, n_drop)

    def query(self, n, embeddings=None):
        self.check_query_num(n)
        _path_list = np.array(self.path_mapping)

        if self.n_drop:
            sigma_c = np.std(embeddings, axis=0)
            _uncertainties = np.mean(sigma_c, axis=-1)
            _uncertainties_sorted = np.argsort(-np.array(_uncertainties))
            return _path_list[_uncertainties_sorted[:int(n)]]
        else:
            raise ValueError("strategy {BayesianDisagreement} requires to set the parameter {n_drop} > 1")


class VarRatioSampling(Strategy):
    """
    Variation Ratios Sampling.
    Reference: Elementary applied statistics: for students in behavioral science. New York: Wiley, 1965
    """

    def __init__(self, path_mapping, n_drop):
        super(VarRatioSampling, self).__init__(path_mapping, n_drop)

    def query(self, n, embeddings=None):
        self.check_query_num(n)
        _path_list = np.array(self.path_mapping)

        if self.n_drop:
            embeddings = embeddings.mean(0)

        _preds = np.max(embeddings, axis=1)
        _uncertainties = np.argsort(-np.array(1.0 - _preds))
        return _path_list[_uncertainties[:int(n)]]



"""
The active learning executor of PyTorch models.
@author huangyz0918 (huangyz0918@gmail.com)
@date 28/05/2022
"""
import copy

import torch
import torch.jit
import torch.nn.functional as F

import os
import warnings
import importlib
import numpy as np
from pathlib import Path
from multiprocessing.pool import ThreadPool
from jina import Executor, Document, DocumentArray, requests

from alaas.server.util import DBManager
from alaas.types import ALStrategyType, ModalityType
from alaas.server.preprocessor import img_transform


class TorchALWorker(Executor):
    """
    TorchALWorker: the backend worker class of PyTorch models.
    """

    def __init__(
            self,
            model_name: str = 'resnet18',
            model_repo: str = 'pytorch/vision:v0.10.0',
            device: str = None,
            minibatch_size: int = 8,
            num_worker_preprocess: int = 4,
            data_home: str = None,
            tokenizer_model: str = None,
            task: str = None,
            transform=None,
            strategy: str = ALStrategyType.LEAST_CONFIDENCE.value,
            *args,
            **kwargs,
    ):
        """
        The parameters of build a TorchWorker.
        @param model_name: the active learning model name (default is resnet18).
        @param model_repo: the model hub name, for example: pytorch/vision:v0.10.0, models will be downloaded from here.
        @param device: the model running device (default is 'cpu').
        @param minibatch_size: the mini batch size for data processing.
        @param num_worker_preprocess: the number of backend workers (multi-thread).
        @param data_home: the data home path for storing the unlabeled data.
        @param tokenizer_model: [for nlp tasks], the model name of tokenizer, for example, 'bert-large-uncased'.
        @param task: [for nlp tasks], the task name of transformers pipeline, for example, 'text-classification'.
        @param transform: the data transform for pre-processing.
        @param strategy: the active learning strategy (for example, 'LeastConfidence').
        @param args: the args for backend server.
        @param kwargs: the kwargs for backend server.
        """
        super().__init__(*args, **kwargs)

        self._strategy = strategy
        self._transform = transform
        self._minibatch_size = minibatch_size
        self._thread_pool = ThreadPool(processes=num_worker_preprocess)

        self._task = task
        self._model_repo = model_repo
        self._model_name = model_name
        self._tokenizer_model = tokenizer_model

        if data_home is None:
            self._data_home = str(Path.home()) + "/.alaas/"
            Path(self._data_home).mkdir(parents=True, exist_ok=True)

        self._db_manager = DBManager(self._data_home + 'cache.db')
        self._data_pool = self._db_manager.read_records()
        self._data_processed = self._db_manager.get_rows()
        self._data_not_processed = self._db_manager.get_rows(inferred=False)

        if not device:
            self._device = 'cuda' if torch.cuda.is_available() else 'cpu'
        else:
            self._device = device

        if not self._device.startswith('cuda') and (
                'OMP_NUM_THREADS' not in os.environ
                and hasattr(self.runtime_args, 'replicas')
        ):
            replicas = getattr(self.runtime_args, 'replicas', 1)
            num_threads = max(1, torch.get_num_threads() // replicas)
            if num_threads < 2:
                warnings.warn(
                    f'Too many replicas ({replicas}) vs too few threads {num_threads} may result in '
                    f'sub-optimal performance.'
                )

                torch.set_num_threads(max(num_threads, 1))
                torch.set_num_interop_threads(1)
        
        # skip loading models while using random sampling strategy.
        if self._strategy != ALStrategyType.RANDOM_SAMPLING.value:
            self._init_model()
        else:
            print("\n skip initializing deep learning models for random sampling strategy. \n")

    def _init_model(self):
        # set up the active learning (for query only) model.
        if self._tokenizer_model and 'huggingface' in self._model_repo:
            self._data_modality = ModalityType.TEXT
            from transformers import AutoTokenizer, pipeline
            _tokenizer = AutoTokenizer.from_pretrained(self._tokenizer_model)
            # build task-specify parameters (TODO: adapt to more tasks).
            hf_extra_parameters = {}
            if self._task == 'text-classification':
                hf_extra_parameters.update({'return_all_scores': True})
            # build the pipeline.
            self._model = pipeline(self._task,
                                model=self._model_name,
                                tokenizer=_tokenizer,
                                device=self._convert_torch_device(),
                                **hf_extra_parameters)
        else:
            self._data_modality = ModalityType.IMAGE
            self._model = torch.hub.load(self._model_repo, model=self._model_name, pretrained=True)
            self._model.eval().to(self._device)

    def _convert_torch_device(self):
        """
        converting the pytorch devices to the format of transformers.
        @return: the processed device.
        """
        if self._device == 'cpu':
            return -1
        elif self._device == 'cuda':
            return 0
        else:
            return self._device.split(':')[-1]

    def _set_input_modality(self, sample: Document):
        """
        Check and set the global data modality for the worker.
        @param sample: one of the sample from all the unlabeled data samples.
        @return: None.
        """
        if sample.blob or sample.uri:
            self._data_modality = ModalityType.IMAGE
        elif sample.text:
            self._data_modality = ModalityType.TEXT
        else:
            raise TypeError("unsupported data modality, data should be neither image or text.")

    def _get_model_embeddings(self, docs: DocumentArray):
        """
        Get the processed data embedding by given deep learning models.
        @param docs: the input data in DocumentArray.
        @return: the return values in DocumentArray with embeddings.
        """
        index_pths = []
        for minibatch, index_list, batch_data in docs.map_batch(
                self._preproc_data,
                batch_size=self._minibatch_size,
                pool=self._thread_pool,
        ):
            index_pths += index_list
            if self._strategy != ALStrategyType.RANDOM_SAMPLING.value:
                if self._data_modality == ModalityType.IMAGE:
                    minibatch.embeddings = (
                        F.softmax(self._model(batch_data.to(self._device)), dim=1)
                            .cpu()
                            .numpy()
                            .astype(np.float32)
                    )
                elif self._data_modality == ModalityType.TEXT:
                    results = []
                    for item in self._model(batch_data):
                        results.append([x['score'] for x in item])
                    minibatch.embeddings = (
                        np.array(results)
                    )
        return index_pths, docs

    @requests(on='/push')
    def push(self, docs: DocumentArray, parameters, **kwargs):
        """
        Active learning push function. This function is designed for large-scale input data/uris,
        the data will be fetched first and store temporary in the centre AL server.
        @param docs: the uploaded/targeted data, can be blob/uri for CV tasks or text for NLP tasks.
        @param parameters: the parameters include the active learning budget, etc.
        @param kwargs: the kwargs for the backend server.
        @return: the queried data in uris/blobs/texts.
        """
        index_pths = []

        # save the docs into local files.
        if parameters['save_path']:
            save_path = parameters['save_path']
        else:
            save_path = self._data_home

        for doc in docs:
            data_path = save_path + doc.id
            index_pths.append(Document(content=data_path))

            if doc.blob:
                doc.save_blob_to_file(data_path)

            elif doc.uri:
                doc.load_uri_to_blob()
                doc.save_blob_to_file(data_path)

            elif doc.text:
                doc.convert_text_to_datauri()
                doc.save_uri_to_file(data_path)

            self._db_manager.insert_record(data_path)

        return DocumentArray(index_pths)

    @requests(on='/query')
    def query(self, docs: DocumentArray, parameters, **kwargs):
        """
        Active learning query function. This function is an end-to-end data query function,
        the data need to be uploaded/downloaded after calling the function.
        @param docs: the uploaded/targeted data, can be blob/uri for CV tasks or text for NLP tasks.
        @param parameters: the parameters include the active learning budget, etc.
            @param budget: the active learning query budget, required for each query.
            @param n_drop: the number of dropout, default is None.
        @param kwargs: the kwargs for the backend server.
        @return: the queried data in uris/blobs/texts.
        """
        index_pths = []
        input_embeddings = None
        with self.monitor(
                name='query_inputs',
                documentation='data download and preprocess time in seconds',
        ):
            with torch.inference_mode():
                if 'n_drop' in parameters.keys() and parameters['n_drop'] is not None:
                    for _ in range(int(parameters['n_drop'])):
                        index_pths, round_docs = self._get_model_embeddings(copy.deepcopy(docs))
                        if input_embeddings is None:
                            input_embeddings = round_docs.embeddings
                            input_embeddings = np.stack((input_embeddings, round_docs.embeddings))
                        else:
                            input_embeddings = np.concatenate((input_embeddings, [round_docs.embeddings]), axis=0)
                else:
                    index_pths, docs = self._get_model_embeddings(docs)
                    input_embeddings = docs.embeddings

            _doc_list = []
            al_method = getattr(importlib.import_module('alaas.server.strategy'), self._strategy)
            if 'n_drop' in parameters.keys() and parameters['n_drop'] is not None:
                al_learner = al_method(path_mapping=index_pths, n_drop=int(parameters['n_drop']))
            else:
                al_learner = al_method(path_mapping=index_pths, n_drop=None)
            query_results = al_learner.query(parameters['budget'], input_embeddings)

            for result in query_results:
                if self._data_modality == ModalityType.IMAGE:
                    _doc_list.append(Document(blob=result.blob, uri=result.uri))

                elif self._data_modality == ModalityType.TEXT:
                    _doc_list.append(Document(text=result.text, uri=result.uri))

            return DocumentArray(_doc_list)

    def _preproc_data(self, docs: DocumentArray):
        """
        pre-processing function for the active learning input data.
        @param docs: the input data.
        @return: original data objects, data index list (for active learning mapping), and the processed data.
        """
        with self.monitor(
                name='preprocess_data_seconds',
                documentation='images preprocess time in seconds',
        ):
            _tensor_list = []
            index_list = []

            self._set_input_modality(docs[0])

            if self._data_modality == ModalityType.IMAGE:
                self._transform = img_transform
                for data in docs:
                    if data.blob:
                        index_doc = Document(blob=data.blob)
                        data.convert_blob_to_image_tensor()
                    elif data.tensor is None and data.uri:
                        # in case user uses HTTP protocol and send data via curl not using .blob (base64), but in .uri
                        index_doc = Document(uri=data.uri)
                        data.load_uri_to_image_tensor()

                    index_list.append(index_doc)
                    _tensor_list.append(self._transform(data.tensor))
                return_outputs = torch.stack(_tensor_list, dim=0)
            elif self._data_modality == ModalityType.TEXT:
                for data in docs:
                    if self._tokenizer_model:
                        index_list.append(Document(text=data.text))
                        # TODO: text data pre-processing here.
                        _tensor_list.append(data.text)
                        return_outputs = _tensor_list
                    else:
                        raise ValueError(
                            """
                            For text data, you need to set the tokenizer model in your 
                            configuration file (e.g., tokenizer: \"distilbert-base-uncased\") 
                            under the \"active_learning.strategy.model\"
                            """)

            return docs, index_list, return_outputs



"""
Clustering based active learning methods.
@author huangyz0918 (huangyz0918@gmail.com)
@date 18/07/2022
"""

import numpy as np
from sklearn.cluster import KMeans

from .base import Strategy


class KMeansSampling(Strategy):
    """
    KMeans Cluster Sampling Method.
    """

    def __init__(self, path_mapping, n_drop):
        super(KMeansSampling, self).__init__(path_mapping, n_drop)

    def query(self, n, embeddings=None):
        self.check_query_num(n)
        _path_list = np.array(self.path_mapping)

        if self.n_drop:
            embeddings = embeddings.mean(0)

        kmeans = KMeans(n_clusters=n)
        kmeans.fit(embeddings)
        distance = kmeans.transform(embeddings)

        query_ids = np.argmin(distance, axis=0)
        return _path_list[query_ids]

# class KCenterGreedy(Strategy):
#     """
#     K-Center Greedy Method.
#     """
#
#     def __init__(self, path_mapping, n_drop):
#         super(KCenterGreedy, self).__init__(path_mapping, n_drop)
#
#     def query(self, n, embeddings=None):
#         self.check_query_num(n)
#         _path_list = np.array(self.path_mapping)
#
#         dist_mat = np.matmul(embeddings, embeddings.transpose())
#         sq = np.array(dist_mat.diagonal()).reshape(len(labeled_idxs), 1)
#         dist_mat *= -2
#         dist_mat += sq
#         dist_mat += sq.transpose()
#         dist_mat = np.sqrt(dist_mat)
#
#         mat = dist_mat[~labeled_idxs, :][:, labeled_idxs]
#
#         for i in range(n):
#             mat_min = mat.min(axis=1)
#             q_idx_ = mat_min.argmax()
#             q_idx = np.arange(self.pool_size)[~labeled_idxs][q_idx_]
#             labeled_idxs[q_idx] = True
#             mat = np.delete(mat, q_idx_, 0)
#             mat = np.append(mat, dist_mat[~labeled_idxs, q_idx][:, None], axis=1)
#
#         return _path_list[np.arange(self.pool_size)[(self.dataset.labeled_idxs ^ labeled_idxs)]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Author: Li Yuanming
Author: huangyz0918
Email: yuanmingleee@gmail.com
Email: huangyz0918@gmail.com
Date: May 23, 2022
"""
from pydantic import BaseModel, BaseSettings

from alaas.types.models.al_strategy import ALStrategyConfigUnion


class ALWorkerConfig(BaseModel):
    protocol: str
    host: str
    port: int
    replicas: int


class ALConfig(BaseModel):
    strategy: ALStrategyConfigUnion
    al_worker: ALWorkerConfig


class Config(BaseSettings):
    name: str
    version: str
    active_learning: ALConfig



"""
Configuration file manager for active learning as a service.
@author: huangyz0918 (huangyz0918@gmail.com)
@author: Li Yuanming
@date: 01/05/2022
"""

import yaml

from alaas.types.models.config import Config


class ConfigManager:

    def __init__(self, config_path):
        self.config: Config = ConfigManager.load_config(config_path)

    @staticmethod
    def load_config(config_path):
        with open(config_path) as f:
            config_obj = yaml.safe_load(f)
        return Config.parse_obj(config_obj)

    @property
    def name(self):
        return self.config.name

    @property
    def version(self):
        return self.config.version

    @property
    def strategy(self):
        return self.config.active_learning.strategy

    @property
    def al_worker(self):
        return self.config.active_learning.al_worker


from .random import RandomSampling
from .uncertainty import LeastConfidence, LeastConfidence, MarginConfidence, RatioConfidence, EntropySampling, \
    BayesianDisagreement, MeanSTDSampling, VarRatioSampling
from .cluster import KMeansSampling
from .diversity import BadgeSampling



"""
Base active learning class.
@author huangyz0918 (huangyz0918@gmail.com)
@date 20/07/2022
"""


class Strategy:
    """The basic active learning strategy class"""

    def __init__(self, path_mapping, n_drop=None):
        """
        @param pool_size: if pool size = 1, the settings is stream-based.
        @param path_mapping: the mapping with input uris.
        @param n_drop: the number of dropout, default is None.
        """
        self.path_mapping = path_mapping
        self.pool_size = len(path_mapping)
        self.n_drop = n_drop

    def check_query_num(self, number):
        """
        Check the input query number is valid or not.
        """
        if number > self.pool_size:
            raise ValueError(f"please indicate a query number smaller than data pool size: {self.pool_size}")

    def query(self, number, embeddings):
        """
        The query function that calls the specific active learning strategy.
        :param number: the query number.
        :@param embeddings: the data embedding.
        :return: the data ids in a numpy array.
        """
        pass



from enum import Enum
from typing import List, Any, Dict, Optional
from pydantic import Field, PositiveInt, BaseModel


class ModalityType(Enum):
    IMAGE = 'MODALITY_IMAGE'
    TEXT = 'MODALITY_TEXT'


class ALModelBase(BaseModel):
    name: str
    batch_size: PositiveInt
    hub: str
    device: str
    tokenizer: Optional[str] = None
    task: Optional[str] = None

    args: List[str] = Field(default_factory=list)
    kwargs: Dict[str, Any] = Field({'pretrained': True})



"""
The Client of ALaaS.
@author huangyz0918 (huangyz0918@gmail.com)
@date 28/05/2022
"""

from urllib.parse import urlparse
from docarray import Document, DocumentArray

from alaas.client.utils import chunks


class Client:
    def __init__(self, server):
        """
        Server scheme is in the format of `scheme://netloc:port`, where
            - scheme: one of grpc, websocket, http, grpcs, websockets, https
            - netloc: the server ip address or hostname
            - port: the public port of the server
        :param server: the server URI
        """
        try:
            r = urlparse(server)
            _port = r.port
            _scheme = r.scheme
            if not _scheme:
                raise
        except:
            raise ValueError(f'{server} is not a valid scheme')

        _tls = False

        if _scheme in ('grpcs', 'https', 'wss'):
            _scheme = _scheme[:-1]
            _tls = True

        if _scheme == 'ws':
            _scheme = 'websocket'  # temp fix for the core

        if _scheme in ('grpc', 'http', 'websocket'):
            _kwargs = dict(host=r.hostname, port=_port, protocol=_scheme, tls=_tls)

            from jina import Client

            self._client = Client(**_kwargs)
            self._async_client = Client(**_kwargs, asyncio=True)
        else:
            raise ValueError(f'{server} is not a valid scheme')

    def query_by_uri(self, input_uris, budget, divide_mode=False, n_drop=None):
        """
        Query the active learner by given a list of image data uris.
        @param input_uris: the input data uris (path).
        @param budget: the querying budget.
        @param divide_mode: query the data by the batch mode, can be applied to large data.
        @param n_drop: the number of dropout, default is None (1).
        @return: queried data uris.
        """
        if divide_mode:
            # set batch_size = 1 for streaming AL settings.
            for batch in chunks(input_uris, 1):
                _temp_list = []
                for uri in batch:
                    _temp_list.append(Document(uri=uri))
                _bth_push_res = self._client.post('/push', DocumentArray(_temp_list), parameters={'save_path': None})
        else:
            _doc_list = []
            for uri in input_uris:
                _doc_list.append(Document(uri=uri))

            response = self._client.post('/query', DocumentArray(_doc_list), request_size=len(_doc_list),
                                         parameters={'budget': budget, 'n_drop': n_drop}).to_list()
            return [x["uri"] for x in response]

        return None

    def query_by_text(self, texts, budget, divide_mode=False, n_drop=None):
        """
        Query the active learner by given a list of text data uris.
        @param texts: the input data texts.
        @param budget: the querying budget.
        @param divide_mode: query the data by the batch mode, can be applied to large data.
        @param n_drop: the number of dropout, default is None (1).
        @return: queried data uris.
        """
        if divide_mode:
            # set batch_size = 1 for streaming AL settings.
            for batch in chunks(texts, 1):
                _temp_list = []
                for txt in batch:
                    _temp_list.append(Document(text=txt, mime_type='text'))
                _bth_push_res = self._client.post('/push', DocumentArray(_temp_list), parameters={'save_path': None})
        else:
            _doc_list = []
            for txt in texts:
                _doc_list.append(Document(text=txt, mime_type='text'))

            response = self._client.post('/query', DocumentArray(_doc_list), request_size=len(_doc_list),
                                         parameters={'budget': budget, 'n_drop': n_drop}).to_list()
            return [x["text"] for x in response]

        return None

    def query_by_images(self, path_list, budget, divide_mode=False, n_drop=None):
        """
        Query the active learner by given a list of text data uris.
        @param path_list: the input data images (local path).
        @param budget: the querying budget.
        @param divide_mode: query the data by the batch mode, can be applied to large data.
        @param n_drop: the number of dropout, default is None (1).
        @return: queried data uris.
        """
        if divide_mode:
            # set batch_size = 1 for streaming AL settings.
            for batch in chunks(path_list, 1):
                _temp_list = []
                for pth in batch:
                    _temp_list.append(Document(blob=open(pth, "rb").read(), mime_type='image'))
                _bth_push_res = self._client.post('/push', DocumentArray(_temp_list), parameters={'save_path': None})
        else:
            _doc_list = []
            for pth in path_list:
                _doc_list.append(Document(blob=open(pth, "rb").read(), mime_type='image'))

            response = self._client.post('/query', DocumentArray(_doc_list), request_size=len(_doc_list),
                                         parameters={'budget': budget, 'n_drop': n_drop}).to_list()
            return [x["blob"] for x in response]

        return None



"""
Diversity based active learning methods.
@author huangyz0918 (huangyz0918@gmail.com)
@date 02/08/2022
"""

import numpy as np
from .base import Strategy


class BadgeSampling(Strategy):
    """
    TODO: to build.
    Reference: Deep Batch Active Learning by Diverse, Uncertain Gradient Lower Bounds (https://arxiv.org/abs/1906.03671).
    @inproceedings{ash2019deep,
      author    = {Jordan T. Ash and
                   Chicheng Zhang and
                   Akshay Krishnamurthy and
                   John Langford and
                   Alekh Agarwal},
      title     = {Deep Batch Active Learning by Diverse, Uncertain Gradient Lower Bounds},
      booktitle = {8th International Conference on Learning Representations, {ICLR} 2020,
                   Addis Ababa, Ethiopia, April 26-30, 2020},
      publisher = {OpenReview.net},
      year      = {2020}
    }
    """

    def __init__(self, path_mapping, n_drop):
        super(BadgeSampling, self).__init__(path_mapping, n_drop)

    def query(self, n, embeddings=None):
        self.check_query_num(n)
        _path_list = np.array(self.path_mapping)
        if self.n_drop:
            embeddings = embeddings.mean(0)
        _uncertainties = np.amax(embeddings, axis=1)
        return _path_list[_uncertainties.argsort()[:int(n)]]
